import logging
import os
import shutil
import subprocess
import tempfile
from typing import Any, Dict, List, Optional, Tuple, Union
from urllib.parse import urlparse

from langchain_text_splitters import Language
from openai import BaseModel

from whiskerrag_types.interface.loader_interface import BaseLoader
from whiskerrag_types.model.knowledge import (
    Knowledge,
    KnowledgeSourceEnum,
    KnowledgeTypeEnum,
)
from whiskerrag_types.model.knowledge_source import GithubRepoSourceConfig
from whiskerrag_types.model.multi_modal import Text
from whiskerrag_types.model.splitter import (
    BaseCodeSplitConfig,
    BaseSplitConfig,
    JSONSplitConfig,
    MarkdownSplitConfig,
    PDFSplitConfig,
    TextSplitConfig,
)
from whiskerrag_utils.loader.file_pattern_manager import FilePatternManager
from whiskerrag_utils.registry import RegisterTypeEnum, register

logger = logging.getLogger(__name__)


class GitFileElementType(BaseModel):
    content: str
    path: str
    mode: str
    url: str
    branch: str
    repo_name: str
    size: int
    sha: str
    position: dict = {}  # VSCode position information


def _check_git_installation() -> bool:
    """check git installation"""
    try:
        subprocess.run(["git", "--version"], check=True, capture_output=True, text=True)
        return True
    except (subprocess.SubprocessError, FileNotFoundError):
        return False


def _get_temp_git_env() -> Dict[str, str]:
    """
    get temporary git environment configuration

    Returns:
        Dict[str, str]: git config
    """
    return {
        "GIT_AUTHOR_NAME": "temp",
        "GIT_AUTHOR_EMAIL": "temp@example.com",
        "GIT_COMMITTER_NAME": "temp",
        "GIT_COMMITTER_EMAIL": "temp@example.com",
        "GIT_TERMINAL_PROMPT": "0",  # disable interactive prompt
    }


def _lazy_import_git() -> Tuple[Any, Any, Any]:
    """Lazy import git modules to avoid direct dependency"""
    try:
        from git import Repo
        from git.exc import GitCommandNotFound, InvalidGitRepositoryError

        return Repo, GitCommandNotFound, InvalidGitRepositoryError
    except ImportError as e:
        raise ImportError(
            "GitPython is required for Git repository loading. "
            "Please install it with: pip install GitPython"
        ) from e


@register(RegisterTypeEnum.KNOWLEDGE_LOADER, KnowledgeSourceEnum.GITHUB_REPO)
class GithubRepoLoader(BaseLoader):
    repo_name: str
    branch_name: Optional[str] = None
    token: Optional[str] = None
    local_repo: Optional[Any] = (
        None  # Use Any instead of object to avoid mypy attr-defined errors
    )
    knowledge: Knowledge

    @property
    def repos_dir(self) -> str:
        """get runtime root folder path"""
        # use system tempdir instead of current dir
        return os.path.join(tempfile.gettempdir(), "repo_download")

    def __init__(
        self,
        knowledge: Knowledge,
    ):
        """
        init GithubRepoLoader

        Args:
            knowledge: Knowledge instance,which include repo source config

        Raises:
            ValueError: invalid repo config
        """
        if not isinstance(knowledge.source_config, GithubRepoSourceConfig):
            raise ValueError("source_config should be GithubRepoSourceConfig")

        self.knowledge = knowledge
        self.repo_name = knowledge.source_config.repo_name
        self.branch_name = knowledge.source_config.branch
        self.token = knowledge.source_config.auth_info
        self.base_url = knowledge.source_config.url.rstrip("/")

        os.makedirs(self.repos_dir, exist_ok=True)

        self.repo_path: Optional[str] = os.path.join(
            self.repos_dir, self.repo_name.replace("/", "_")
        )

        try:
            self._load_repo()
        except Exception as e:
            logger.error(f"Failed to load repo: {e}")
            raise ValueError(
                f"Failed to load repo {self.repo_name} with branch "
                f"{self.branch_name}. Error: {str(e)}"
            )

    def _load_repo(self) -> None:
        Repo, GitCommandNotFound, InvalidGitRepositoryError = _lazy_import_git()

        if not _check_git_installation():
            raise ValueError("Git is not installed in the system")

        if not self.repo_path:
            raise ValueError("Repository path not initialized")

        try:
            parsed_url = urlparse(self.base_url)
            if not parsed_url.scheme or parsed_url.scheme != "https":
                raise ValueError(
                    f"Invalid URL scheme: {self.base_url}. "
                    "Only HTTPS URLs are supported"
                )

            clone_url = f"{self.base_url}/{self.repo_name}.git"

            git_env = _get_temp_git_env()

            if self.token:
                git_env.update(
                    {
                        "GIT_ASKPASS": "echo",
                        "GIT_USERNAME": "git",
                        "GIT_PASSWORD": self.token,
                    }
                )

            if os.path.exists(self.repo_path):
                try:
                    self.local_repo = Repo(self.repo_path)
                    self._update_repo()
                except Exception as e:
                    logger.warning(f"Failed to update existing repo: {e}")
                    shutil.rmtree(self.repo_path)
                    self._clone_repo(clone_url, git_env)
            else:
                self._clone_repo(clone_url, git_env)

            if not self.branch_name and self.local_repo:
                try:
                    self.branch_name = self.local_repo.active_branch.name
                except Exception as e:
                    logger.warning(f"Failed to get default branch name: {str(e)}")
                    self.branch_name = "main"

            logger.info(f"Successfully loaded repository at {self.repo_path}")

        except Exception as e:
            raise ValueError(
                f"Failed to load repo {self.repo_name} with branch "
                f"{self.branch_name}. Error: {str(e)}"
            )

    def _clone_repo(self, clone_url: str, git_env: Dict[str, str]) -> None:
        Repo, GitCommandNotFound, InvalidGitRepositoryError = _lazy_import_git()

        if not self.repo_path:
            raise ValueError("Repository path not initialized")

        try:
            if self.branch_name:
                self.local_repo = Repo.clone_from(
                    url=clone_url,
                    to_path=self.repo_path,
                    depth=1,
                    single_branch=True,
                    env=git_env,
                    branch=self.branch_name,
                )
            else:
                self.local_repo = Repo.clone_from(
                    url=clone_url,
                    to_path=self.repo_path,
                    depth=1,
                    single_branch=True,
                    env=git_env,
                )
        except GitCommandNotFound:
            raise ValueError("Git command not found. Please ensure git is installed.")
        except InvalidGitRepositoryError as e:
            raise ValueError(f"Invalid git repository: {str(e)}")
        except Exception as e:
            if "Authentication failed" in str(e):
                raise ValueError("Authentication failed. Please check your token.")
            raise

    def _update_repo(self) -> None:
        # Ensure git is available before proceeding
        if not _check_git_installation():
            raise ValueError("Git is not installed in the system")

        if not self.local_repo:
            raise ValueError("Repository not initialized")

        if self.branch_name:
            self.local_repo.git.checkout(self.branch_name)
        self.local_repo.remotes.origin.pull()

    @staticmethod
    def get_knowledge_type_by_ext(ext: str) -> KnowledgeTypeEnum:
        ext = ext.lower()
        ext_to_type = {
            ".md": KnowledgeTypeEnum.MARKDOWN,
            ".mdx": KnowledgeTypeEnum.MARKDOWN,
            ".txt": KnowledgeTypeEnum.TEXT,
            ".json": KnowledgeTypeEnum.JSON,
            ".pdf": KnowledgeTypeEnum.PDF,
            ".docx": KnowledgeTypeEnum.DOCX,
            ".rst": KnowledgeTypeEnum.RST,
            ".py": KnowledgeTypeEnum.PYTHON,
            ".js": KnowledgeTypeEnum.JS,
            ".ts": KnowledgeTypeEnum.TS,
            ".go": KnowledgeTypeEnum.GO,
            ".java": KnowledgeTypeEnum.JAVA,
            ".cpp": KnowledgeTypeEnum.CPP,
            ".c": KnowledgeTypeEnum.C,
            ".h": KnowledgeTypeEnum.C,
            ".hpp": KnowledgeTypeEnum.CPP,
            ".cs": KnowledgeTypeEnum.CSHARP,
            ".kt": KnowledgeTypeEnum.KOTLIN,
            ".swift": KnowledgeTypeEnum.SWIFT,
            ".php": KnowledgeTypeEnum.PHP,
            ".rb": KnowledgeTypeEnum.RUBY,
            ".rs": KnowledgeTypeEnum.RUST,
            ".scala": KnowledgeTypeEnum.SCALA,
            ".sol": KnowledgeTypeEnum.SOL,
            ".html": KnowledgeTypeEnum.HTML,
            ".css": KnowledgeTypeEnum.TEXT,
            ".lua": KnowledgeTypeEnum.LUA,
            ".m": KnowledgeTypeEnum.TEXT,  # Objective-C/MATLAB等
            ".sh": KnowledgeTypeEnum.TEXT,
            ".yml": KnowledgeTypeEnum.TEXT,
            ".yaml": KnowledgeTypeEnum.TEXT,
            ".tex": KnowledgeTypeEnum.LATEX,
            ".jpg": KnowledgeTypeEnum.IMAGE,
            ".jpeg": KnowledgeTypeEnum.IMAGE,
            ".png": KnowledgeTypeEnum.IMAGE,
            ".gif": KnowledgeTypeEnum.IMAGE,
            ".bmp": KnowledgeTypeEnum.IMAGE,
            ".svg": KnowledgeTypeEnum.IMAGE,
        }
        return ext_to_type.get(ext, KnowledgeTypeEnum.TEXT)

    def _get_split_config_for_knowledge_type(
        self, knowledge_type: KnowledgeTypeEnum
    ) -> Union[
        MarkdownSplitConfig,
        JSONSplitConfig,
        PDFSplitConfig,
        BaseCodeSplitConfig,
        TextSplitConfig,
        BaseSplitConfig,
    ]:
        """
        Generate appropriate split_config based on knowledge type

        Args:
            knowledge_type: The type of knowledge

        Returns:
            Appropriate split configuration for the knowledge type
        """
        # Use default chunk_size and chunk_overlap from GithubRepoParseConfig
        default_chunk_size = 1500
        default_chunk_overlap = 200

        if knowledge_type == KnowledgeTypeEnum.MARKDOWN:
            return MarkdownSplitConfig(
                chunk_size=default_chunk_size,
                chunk_overlap=default_chunk_overlap,
                separators=[
                    "\n#{1,6} ",
                    "```\n",
                    "\n\\*\\*\\*+\n",
                    "\n---+\n",
                    "\n___+\n",
                    "\n\n",
                    "\n",
                    " ",
                    "",
                ],
                is_separator_regex=True,
                keep_separator="start",
                extract_header_first=True,
            )
        elif knowledge_type == KnowledgeTypeEnum.JSON:
            return JSONSplitConfig(
                max_chunk_size=default_chunk_size,
                min_chunk_size=min(200, default_chunk_size - 200),
            )
        elif knowledge_type == KnowledgeTypeEnum.PDF:
            return PDFSplitConfig(
                chunk_size=default_chunk_size,
                chunk_overlap=default_chunk_overlap,
                extract_images=False,
                table_extract_mode="text",
            )
        elif knowledge_type in [
            KnowledgeTypeEnum.PYTHON,
            KnowledgeTypeEnum.JS,
            KnowledgeTypeEnum.TS,
            KnowledgeTypeEnum.GO,
            KnowledgeTypeEnum.JAVA,
            KnowledgeTypeEnum.CPP,
            KnowledgeTypeEnum.C,
            KnowledgeTypeEnum.CSHARP,
            KnowledgeTypeEnum.KOTLIN,
            KnowledgeTypeEnum.SWIFT,
            KnowledgeTypeEnum.PHP,
            KnowledgeTypeEnum.RUBY,
            KnowledgeTypeEnum.RUST,
            KnowledgeTypeEnum.SCALA,
            KnowledgeTypeEnum.SOL,
            KnowledgeTypeEnum.LUA,
        ]:
            # Map knowledge types to Language enum
            language_map = {
                KnowledgeTypeEnum.PYTHON: Language.PYTHON,
                KnowledgeTypeEnum.JS: Language.JS,
                KnowledgeTypeEnum.TS: Language.TS,
                KnowledgeTypeEnum.GO: Language.GO,
                KnowledgeTypeEnum.JAVA: Language.JAVA,
                KnowledgeTypeEnum.CPP: Language.CPP,
                KnowledgeTypeEnum.C: Language.C,
                KnowledgeTypeEnum.CSHARP: Language.CSHARP,
                KnowledgeTypeEnum.KOTLIN: Language.KOTLIN,
                KnowledgeTypeEnum.SWIFT: Language.SWIFT,
                KnowledgeTypeEnum.PHP: Language.PHP,
                KnowledgeTypeEnum.RUBY: Language.RUBY,
                KnowledgeTypeEnum.RUST: Language.RUST,
                KnowledgeTypeEnum.SCALA: Language.SCALA,
                KnowledgeTypeEnum.SOL: Language.SOL,
                KnowledgeTypeEnum.LUA: Language.LUA,
            }
            return BaseCodeSplitConfig(
                language=language_map.get(knowledge_type, Language.MARKDOWN),
                chunk_size=default_chunk_size,
                chunk_overlap=default_chunk_overlap,
            )
        elif knowledge_type == KnowledgeTypeEnum.TEXT:
            return TextSplitConfig(
                chunk_size=default_chunk_size,
                chunk_overlap=default_chunk_overlap,
                separators=["\n\n", "\n", " ", ""],
                is_separator_regex=False,
                keep_separator=False,
            )
        else:
            # For other types (HTML, RST, LATEX, etc.), use BaseSplitConfig
            return BaseSplitConfig(
                chunk_size=default_chunk_size,
                chunk_overlap=default_chunk_overlap,
            )

    def _get_file_position_info(self, file_path: str, relative_path: str) -> dict:
        """
        Get position information that can be used for URL construction and remote jumping

        Args:
            file_path: Full path to the file
            relative_path: Relative path from repo root

        Returns:
            dict: Position information for remote repository jumping
        """
        try:
            with open(file_path, "r", encoding="utf-8", errors="ignore") as f:
                lines = f.readlines()
                total_lines = len(lines)

            return {
                "file_path": relative_path,
                "start_line": 1,
                "end_line": total_lines,
                "total_lines": total_lines,
            }
        except Exception as e:
            logger.warning(
                f"Could not read file for position info {relative_path}: {e}"
            )
            return {
                "file_path": relative_path,
                "start_line": 1,
                "end_line": None,
                "total_lines": None,
            }

    def generate_jump_url(
        self,
        relative_path: str,
        line: Optional[int] = None,
        end_line: Optional[int] = None,
    ) -> str:
        """
        Generate a jump URL for external systems to create clickable links

        Args:
            relative_path: Relative path to the file
            line: Optional line number to jump to
            end_line: Optional end line for range jumping

        Returns:
            str: Complete URL that can be used for jumping to the file/line
        """
        base_url = (
            f"{self.base_url}/{self.repo_name}/blob/{self.branch_name}/{relative_path}"
        )

        if line is not None:
            if end_line is not None and end_line != line:
                return f"{base_url}#L{line}-L{end_line}"
            else:
                return f"{base_url}#L{line}"
        return base_url

    async def decompose(self) -> List[Knowledge]:
        """
        decompose knowledge units in the repository

        Returns:
            List[Knowledge]: knowledge list

        Raises:
            ValueError: when the repository is not properly initialized
        """
        if not self.local_repo or not self.repo_path:
            raise ValueError("Repository not properly initialized")

        # Ensure git is available before proceeding with git operations
        if not _check_git_installation():
            raise ValueError("Git is not installed in the system")

        # count the total file size of the repo
        total_size = 0
        for root, _, files in os.walk(self.repo_path):
            if ".git" in root:
                continue
            for file in files:
                file_path = os.path.join(root, file)
                try:
                    total_size += os.path.getsize(file_path)
                except Exception:
                    continue

        # initialize file pattern manager
        split_config = getattr(self.knowledge, "split_config", None)
        if split_config and getattr(split_config, "type", None) == "github_repo":
            pattern_manager = FilePatternManager(
                config=split_config, repo_path=self.repo_path
            )
            warnings = pattern_manager.validate_patterns()
            if warnings:
                logger.warning(
                    "Pattern configuration warnings:\n" + "\n".join(warnings)
                )
        else:
            # create a compatible configuration dictionary
            dummy_config = {
                "include_patterns": ["*.md", "*.mdx"],
                "ignore_patterns": [],
                "no_gitignore": True,
                "no_default_ignore_patterns": False,
            }
            pattern_manager = FilePatternManager(
                config=dummy_config, repo_path=self.repo_path
            )

        current_commit = self.local_repo.head.commit

        github_repo_list: List[Knowledge] = []

        for root, _, files in os.walk(self.repo_path):
            if ".git" in root:
                continue

            for file in files:
                file_path = os.path.join(root, file)
                relative_path = os.path.relpath(file_path, self.repo_path)

                if not pattern_manager.should_include_file(relative_path):
                    continue

                try:
                    file_size = os.path.getsize(file_path)
                    git_file_path = relative_path.replace("\\", "/")
                    blob = current_commit.tree / git_file_path
                    file_sha = blob.hexsha
                    ext = os.path.splitext(relative_path)[1].lower()
                    knowledge_type = self.get_knowledge_type_by_ext(ext)
                    file_url = (
                        f"{self.base_url}/{self.repo_name}/blob/"
                        f"{self.branch_name}/{relative_path}"
                    )
                    # Generate appropriate split_config for this knowledge type
                    knowledge_split_config = self._get_split_config_for_knowledge_type(
                        knowledge_type
                    )
                    # Get accurate position information
                    position_info = self._get_file_position_info(
                        file_path, relative_path
                    )

                    knowledge = Knowledge(
                        source_type=KnowledgeSourceEnum.GITHUB_FILE,
                        knowledge_type=knowledge_type,
                        knowledge_name=f"{self.repo_name}/{relative_path}",
                        embedding_model_name=self.knowledge.embedding_model_name,
                        source_config={
                            **self.knowledge.source_config.model_dump(),
                            "path": relative_path,
                        },
                        tenant_id=self.knowledge.tenant_id,
                        file_size=file_size,
                        file_sha=file_sha,
                        space_id=self.knowledge.space_id,
                        split_config=knowledge_split_config,
                        parent_id=self.knowledge.knowledge_id,
                        enabled=True,
                        metadata={
                            "_reference_url": file_url,
                            "branch": self.branch_name,
                            "repo_name": self.repo_name,
                            "path": relative_path,
                            # Position information for remote jumping
                            "position": position_info,
                        },
                    )
                    github_repo_list.append(knowledge)

                except Exception as e:
                    logger.warning(f"Error processing file {relative_path}: {e}")
                    continue

        return github_repo_list

    async def load(self) -> List[Text]:
        """
        返回项目的文件目录树结构信息和作者信息，便于大模型理解
        """
        if not self.repo_path:
            raise ValueError("Repository not properly initialized")

        def build_tree(path: str, prefix: str = "") -> str:
            entries = sorted(os.listdir(path))
            tree_lines = []
            for idx, entry in enumerate(entries):
                full_path = os.path.join(path, entry)
                connector = "└── " if idx == len(entries) - 1 else "├── "
                tree_lines.append(f"{prefix}{connector}{entry}")
                if os.path.isdir(full_path) and entry != ".git":
                    extension = "    " if idx == len(entries) - 1 else "│   "
                    tree_lines.append(build_tree(full_path, prefix + extension))
            return "\n".join(tree_lines)

        root_name = os.path.basename(self.repo_path.rstrip(os.sep))
        tree_str = (
            f"repo: {root_name}\n" + "project tree: " + build_tree(self.repo_path)
        )

        try:
            if not self.local_repo:
                raise ValueError("Repository not initialized")

            # Ensure git is available before accessing git objects
            if not _check_git_installation():
                raise ValueError("Git is not installed in the system")

            first_commit = next(
                self.local_repo.iter_commits(
                    rev=self.branch_name, max_count=1, reverse=True
                )
            )
            author_name = first_commit.author.name
            author_email = first_commit.author.email
        except Exception:
            author_name = None
            author_email = None

        return [
            Text(
                content=tree_str,
                metadata={
                    **self.knowledge.metadata,
                    "repo_name": self.repo_name,
                    "author_name": author_name,
                    "author_email": author_email,
                },
            )
        ]

    async def on_load_finished(self) -> None:
        """clean resource"""
        try:
            if self.repo_path and os.path.exists(self.repo_path):
                shutil.rmtree(self.repo_path)
                self.repo_path = None
                self.local_repo = None
                logger.info(f"Cleaned up temporary directory for {self.repo_name}")
        except Exception as e:
            logger.error(f"Error cleaning up resources: {e}")

    async def get_file_by_path(self, path: str) -> GitFileElementType:
        if not self.local_repo:
            raise ValueError("Repository not initialized")

        if not self.repo_path:
            raise ValueError("Repository path not initialized")

        # Ensure git is available before accessing git objects
        if not _check_git_installation():
            raise ValueError("Git is not installed in the system")

        full_path = os.path.join(self.repo_path, path)
        if not os.path.exists(full_path):
            raise ValueError(f"File not found: {path}")

        try:
            blob = self.local_repo.head.commit.tree[path]
            # type check to ensure source_config is GithubRepoSourceConfig
            if not isinstance(self.knowledge.source_config, GithubRepoSourceConfig):
                raise ValueError("Invalid source config type")
            base_url = self.knowledge.source_config.url.rstrip("/")
            file_url = f"{base_url}/{self.repo_name}/blob/{self.branch_name}/{path}"

            # read file content
            try:
                with open(full_path, "r", encoding="utf-8") as f:
                    content = f.read()
            except UnicodeDecodeError:
                # if not a text file, try binary reading and convert to base64
                with open(full_path, "rb") as f:
                    import base64

                    content = base64.b64encode(f.read()).decode("utf-8")

            return GitFileElementType(
                content=content,
                path=path,
                mode=blob.mode_str,
                url=file_url,
                branch=self.branch_name or "main",
                repo_name=self.repo_name,
                size=blob.size,
                sha=blob.hexsha,
            )
        except Exception as e:
            logger.error(f"Error getting file info: {e}")
            raise ValueError(f"Failed to get file info for {path}: {str(e)}")



import base64
from abc import ABC, abstractmethod
from pathlib import Path
from typing import (
    Any,
    AsyncIterator,
    Dict,
    Generic,
    Iterator,
    List,
    Optional,
    TypeVar,
    Union,
)

from pydantic import BaseModel

from whiskerrag_types.model.knowledge import Knowledge

T = TypeVar("T", bound=BaseModel)

ContentType = Union[
    str, Dict[str, Any]
]  # dict用于图片输入，格式为{"type": "image", "data": bytes/str}


class BaseLLM(ABC, Generic[T]):
    """基础大语言模型接口，参考LangChain和OpenAI的设计"""

    def __init__(self, knowledge: Knowledge) -> None:
        self.knowledge = knowledge

    @abstractmethod
    async def chat(self, content: Union[str, List[ContentType]], **kwargs: Any) -> T:
        """
        异步聊天方法，类似于LangChain的invoke

        Args:
            content: 输入内容，可以是字符串或包含文本和图片的列表
            **kwargs: 其他参数，如temperature, max_tokens等

        Returns:
            返回模型响应
        """
        pass

    @abstractmethod
    async def stream_chat(
        self, content: Union[str, List[ContentType]], **kwargs: Any
    ) -> AsyncIterator[T]:
        """
        异步流式聊天方法，类似于LangChain的astream

        Args:
            content: 输入内容，可以是字符串或包含文本和图片的列表
            **kwargs: 其他参数，如temperature, max_tokens等

        Yields:
            流式返回模型响应块
        """
        # 这里需要yield而不是pass，但是为了抽象方法，使用以下方式
        if False:  # pragma: no cover
            yield  # pragma: no cover

    def chat_sync(self, content: Union[str, List[ContentType]], **kwargs: Any) -> T:
        """
        同步聊天方法，类似于LangChain的invoke的同步版本

        Args:
            content: 输入内容，可以是字符串或包含文本和图片的列表
            **kwargs: 其他参数，如temperature, max_tokens等

        Returns:
            返回模型响应
        """
        import asyncio

        # 检查是否在事件循环中
        try:
            asyncio.get_running_loop()
            # 如果在事件循环中，创建任务
            import concurrent.futures

            def run_in_thread() -> T:
                new_loop = asyncio.new_event_loop()
                asyncio.set_event_loop(new_loop)
                try:
                    return new_loop.run_until_complete(self.chat(content, **kwargs))
                finally:
                    new_loop.close()

            with concurrent.futures.ThreadPoolExecutor() as executor:
                future = executor.submit(run_in_thread)
                return future.result()
        except RuntimeError:
            # 没有运行中的事件循环，可以直接调用
            return asyncio.run(self.chat(content, **kwargs))

    def stream_chat_sync(
        self, content: Union[str, List[ContentType]], **kwargs: Any
    ) -> Iterator[T]:
        """
        同步流式聊天方法，类似于LangChain的stream

        Args:
            content: 输入内容，可以是字符串或包含文本和图片的列表
            **kwargs: 其他参数，如temperature, max_tokens等

        Yields:
            流式返回模型响应块
        """
        import asyncio
        import concurrent.futures

        def run_in_thread() -> List[T]:
            new_loop = asyncio.new_event_loop()
            asyncio.set_event_loop(new_loop)
            try:

                async def _run() -> List[T]:
                    results = []
                    async for chunk in self.stream_chat(content, **kwargs):
                        results.append(chunk)
                    return results

                return new_loop.run_until_complete(_run())
            finally:
                new_loop.close()

        # 检查是否在事件循环中
        try:
            asyncio.get_running_loop()
            # 如果在事件循环中，在新线程中运行
            with concurrent.futures.ThreadPoolExecutor() as executor:
                future = executor.submit(run_in_thread)
                results = future.result()
        except RuntimeError:
            # 没有运行中的事件循环，可以直接调用
            results = run_in_thread()

        for result in results:
            yield result

    @staticmethod
    def prepare_image_content(
        image_path: Optional[Union[str, Path]] = None,
        image_data: Optional[bytes] = None,
        image_url: Optional[str] = None,
        image_format: str = "auto",
    ) -> Dict[str, str]:
        """
        准备图片内容的辅助方法

        Args:
            image_path: 图片文件路径
            image_data: 图片二进制数据
            image_url: 图片URL
            image_format: 图片格式，auto表示自动检测

        Returns:
            格式化的图片内容字典
        """
        if image_path:
            image_path = Path(image_path)
            with open(image_path, "rb") as f:
                image_data = f.read()
            if image_format == "auto":
                image_format = image_path.suffix.lower().lstrip(".")

        if image_data:
            if isinstance(image_data, bytes):
                image_b64 = base64.b64encode(image_data).decode("utf-8")
            else:
                image_b64 = image_data

            return {
                "type": "image",
                "image_url": f"data:image/{image_format};base64,{image_b64}",
            }
        elif image_url:
            return {"type": "image", "image_url": image_url}
        else:
            raise ValueError("Must provide either image_path, image_data, or image_url")

    @classmethod
    async def sync_health_check(cls) -> bool:
        """
        同步健康检查方法，用于注册时验证

        Returns:
            True表示健康检查通过，False表示失败
        """
        # 默认实现返回True，子类可以重写进行具体检查
        return True



import asyncio
import logging
from typing import List, Optional, Tuple, TypedDict, Union

from whiskerrag_types.interface.embed_interface import BaseEmbedding
from whiskerrag_types.model.chunk import Chunk
from whiskerrag_types.model.knowledge import Knowledge, KnowledgeTypeEnum
from whiskerrag_types.model.multi_modal import Image, Text

from .registry import RegisterTypeEnum, get_register, init_register, register

logger = logging.getLogger("whisker")


class DiffResult(TypedDict):
    to_add: List[Knowledge]
    to_delete: List[Knowledge]
    unchanged: List[Knowledge]


async def _process_parse_item(
    parse_item: Union[Text, Image],
    knowledge: Knowledge,
    EmbeddingCls: type[BaseEmbedding],
    semaphore: asyncio.Semaphore,
) -> Optional[Chunk]:
    async with semaphore:
        try:
            if isinstance(parse_item, Text):
                embedding = await EmbeddingCls().embed_text(
                    parse_item.content, timeout=30
                )
            elif isinstance(parse_item, Image):
                embedding = await EmbeddingCls().embed_image(parse_item, timeout=30)
            else:
                print(f"[warn]: illegal split item :{parse_item}")
                return None

            combined_metadata = {**knowledge.metadata}
            if isinstance(parse_item, Text) and parse_item.metadata:
                combined_metadata.update(parse_item.metadata)

            # Extract specific fields from metadata according to rules
            # knowledge.metadata._tags, _f1, _f2, _f3, _f4, _f5 -> chunk.tags, chunk.f1, chunk.f2, etc.
            tags = combined_metadata.get("_tags")
            if isinstance(tags, str):
                tags = [tag.strip() for tag in tags.split(",") if tag.strip()]
            elif not isinstance(tags, list):
                tags = None

            return Chunk(
                context=(
                    parse_item.content
                    if isinstance(parse_item, Text)
                    else parse_item.url
                ),
                enabled=knowledge.enabled,
                metadata=combined_metadata,
                # Assign specific fields from metadata
                tags=tags,
                f1=combined_metadata.get("_f1"),
                f2=combined_metadata.get("_f2"),
                f3=combined_metadata.get("_f3"),
                f4=combined_metadata.get("_f4"),
                f5=combined_metadata.get("_f5"),
                embedding=embedding,
                knowledge_id=knowledge.knowledge_id,
                embedding_model_name=knowledge.embedding_model_name,
                space_id=knowledge.space_id,
                tenant_id=knowledge.tenant_id,
            )
        except Exception as e:
            logger.error(f"Error processing parse item: {e}")
            return None


def _get_unique_origin_list(
    origin_list: List[Knowledge],
) -> Tuple[List[Knowledge], List[Knowledge]]:
    to_delete = []
    seen_file_shas = set()
    unique_origin_list = []
    for item in origin_list:
        if item.file_sha not in seen_file_shas:
            seen_file_shas.add(item.file_sha)
            unique_origin_list.append(item)
        else:
            to_delete.append(item)
    return to_delete, unique_origin_list


async def decompose_knowledge(
    knowledge: Knowledge,
    semaphore: Optional[asyncio.Semaphore] = None,
    max_concurrency: int = 4,
) -> List[Knowledge]:
    if semaphore is None:
        semaphore = asyncio.Semaphore(max_concurrency)
    LoaderCls = get_register(RegisterTypeEnum.KNOWLEDGE_LOADER, knowledge.source_type)
    async with semaphore:
        knowledge_list = await LoaderCls(knowledge).decompose()
    if not knowledge_list:
        return []
    tasks = [decompose_knowledge(k, semaphore) for k in knowledge_list]
    results = await asyncio.gather(*tasks)
    flat = [item for sublist in results for item in sublist]
    return flat if flat else knowledge_list


async def get_chunks_by_knowledge(
    knowledge: Knowledge, semaphore_num: int = 4
) -> List[Chunk]:
    """
    Convert knowledge into vectorized chunks with controlled concurrency
    """
    source_type = knowledge.source_type
    knowledge_type = knowledge.knowledge_type
    parse_type = getattr(
        knowledge.split_config,
        "type",
        "base_image" if knowledge_type is KnowledgeTypeEnum.IMAGE else "base_text",
    )
    LoaderCls = get_register(RegisterTypeEnum.KNOWLEDGE_LOADER, source_type)
    ParserCls = get_register(RegisterTypeEnum.PARSER, parse_type)
    EmbeddingCls = get_register(
        RegisterTypeEnum.EMBEDDING, knowledge.embedding_model_name
    )
    # If no parser, return empty list
    if ParserCls is None:
        logger.warn(f"No parser found for type: {parse_type}")
        return []
    # If no embedding model, return empty list
    if EmbeddingCls is None:
        logger.warn(
            f"[warn]: No embedding model found for name: {knowledge.embedding_model_name}"
        )
        return []
    loaded_contents = []
    if LoaderCls is None:
        # If no loader, directly parse the knowledge object itself
        logger.warn(
            f"No loader found for source type: {knowledge.source_type}, attempting to parse knowledge directly."
        )
        parse_results = await ParserCls().parse(knowledge, None)
    else:
        # Use loader to load contents
        loaded_contents = await LoaderCls(knowledge).load()
        if not loaded_contents:
            logger.warn(
                f"Loader returned no content for source type: {knowledge.source_type}."
            )
            return []
        # Parse loaded contents
        parse_results = []
        for content in loaded_contents:
            split_result = await ParserCls().parse(knowledge, content)
            parse_results.extend(split_result)

    semaphore = asyncio.Semaphore(semaphore_num)
    tasks = [
        _process_parse_item(parse_item, knowledge, EmbeddingCls, semaphore)
        for parse_item in parse_results
    ]
    chunks = await asyncio.gather(*tasks)
    return [chunk for chunk in chunks if chunk is not None]


def get_diff_knowledge_by_sha(
    origin_list: Optional[List[Knowledge]] = None,
    new_list: Optional[List[Knowledge]] = None,
) -> DiffResult:
    try:
        origin_list = origin_list or []
        new_list = new_list or []

        to_delete = []
        to_delete_origin, unique_origin_list = _get_unique_origin_list(origin_list)
        to_delete.extend(to_delete_origin)
        _, unique_new_list = _get_unique_origin_list(new_list)

        origin_map = {item.file_sha: item for item in unique_origin_list}

        to_add = []
        unchanged = []
        for new_item in unique_new_list:
            if new_item.file_sha not in origin_map:
                to_add.append(new_item)
            else:
                unchanged.append(new_item)
                del origin_map[new_item.file_sha]

        to_delete.extend(list(origin_map.values()))

        return {"to_add": to_add, "to_delete": to_delete, "unchanged": unchanged}
    except Exception as error:
        logger.error(f"Error in _get_diff_knowledge_by_sha: {error}")
        return {"to_add": [], "to_delete": [], "unchanged": []}


__all__ = [
    "get_register",
    "register",
    "RegisterTypeEnum",
    "init_register",
    "SplitterEnum",
    "decompose_knowledge",
    "get_chunks_by_knowledge",
    "DiffResult",
    "get_diff_knowledge_by_sha",
]



from typing import List, Optional

from whiskerrag_types.interface.loader_interface import BaseLoader
from whiskerrag_types.model import (
    GithubFileSourceConfig,
    Knowledge,
    KnowledgeSourceEnum,
)
from whiskerrag_types.model.multi_modal import Text
from whiskerrag_utils.loader.git_repo_loader import GithubRepoLoader
from whiskerrag_utils.registry import RegisterTypeEnum, register


@register(RegisterTypeEnum.KNOWLEDGE_LOADER, KnowledgeSourceEnum.GITHUB_FILE)
class GithubFileLoader(BaseLoader[Text]):
    """
    Load a file from a GitHub repository.
    """

    knowledge: Knowledge
    path: str
    commit_id: Optional[str]
    repo_loader: GithubRepoLoader

    def __init__(
        self,
        knowledge: Knowledge,
    ):
        self.knowledge = knowledge
        if not isinstance(knowledge.source_config, GithubFileSourceConfig):
            raise ValueError("source_config should be GithubFileSourceConfig")
        self.path = knowledge.source_config.path
        self.repo_loader = GithubRepoLoader(knowledge)

    async def load(self) -> List[Text]:
        file_element = await self.repo_loader.get_file_by_path(self.path)
        return [
            Text(
                content=file_element.content,
                metadata=self.knowledge.metadata,
            )
        ]

    async def decompose(self) -> List[Knowledge]:
        return []

    async def on_load_finished(self) -> None:
        """
        Lifecycle method called when the loading task is finished.
        Subclasses can implement this to perform any cleanup or post-processing.
        """
        pass
